/*
    -- MAGMA (version 1.3.1) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date January 2015

       @generated from zlascl_diag.cpp normal z -> d, Fri Jan 30 18:21:54 2015
*/
//#include "common_magma.h"
#include "magma.h"
#include "magma_d_server.h"

#define NB 64


// each thread block does one NB x n block row of A.
// each thread does one row, starting from left edge and moving right to diagonal.
void
dlascl_diag_lower(int grid, int threads,
                  int m, int n, 
                  magmaDouble_ptr D, int ldd, 
                  magmaDouble_ptr A, int lda)
{
    //#pragma omp parallel for                                                             
    for(int blockIdx=0; blockIdx < grid; blockIdx++) {
        for(int threadIdx=0; threadIdx < threads; threadIdx++) {
            int ind = blockIdx * NB + threadIdx;

            if (ind < m) {
                for(int j=0; j < n; j++ )
                    A[ind + j*lda] /= D[j+j*ldd];
            }
        }
    }
}


// each thread block does one NB x n block row of A.
// each thread does one row, starting from right edge and moving left to diagonal.
void
dlascl_diag_upper(int grid, int threads,
                  int m, int n, 
                  magmaDouble_ptr D, int ldd, 
                  magmaDouble_ptr A, int lda)
{
    //#pragma omp parallel for
    for(int blockIdx=0; blockIdx < grid; blockIdx++) {
        for(int threadIdx=0; threadIdx < threads; threadIdx++) {
            int ind = blockIdx * NB + threadIdx;
            
            if (ind < m) {
                for(int j=0; j < n; j++ )
                    A[ind + j*lda] /= D[ind+ind*ldd];
            }
        }
    }
}

#define max(a, b) ((a) > (b) ? (a) : (b))

/**
    Purpose
    -------
    DLASCL2 scales the M by N real matrix A by the real diagonal matrix dD.
    TYPE specifies that A may be full, upper triangular, lower triangular.

    Arguments
    ---------
    \param[in]
    type    magma_type_t
            TYPE indices the storage type of the input matrix A.
            = MagmaFull:   full matrix.
            = MagmaLower:  lower triangular matrix.
            = MagmaUpper:  upper triangular matrix.
            Other formats that LAPACK supports, MAGMA does not currently support.

    \param[in]
    m       INTEGER
            The number of rows of the matrix A.  M >= 0.

    \param[in]
    n       INTEGER
            The number of columns of the matrix A.  N >= 0.

    \param[in]
    dD      DOUBLE PRECISION vector, dimension (M)
            The diagonal matrix containing the scalar factors. Stored as a vector.

    \param[in,out]
    dA      DOUBLE PRECISION array, dimension (LDDA,N)
            The matrix to be scaled by dD.  See TYPE for the
            storage type.

    \param[in]
    ldda    INTEGER
            The leading dimension of the array A.  LDDA >= max(1,M).

    \param[out]
    info    INTEGER
      -     = 0:  successful exit
      -     < 0:  if INFO = -i, the i-th argument had an illegal value.

    @ingroup magma_daux2
    ********************************************************************/
void call_dlascl_diag(magma_mic_dlascl_diag_param param)
{
    magma_type_t uplo         = param.uplo;
    magma_int_t m             = param.m;
    magma_int_t n             = param.n;
    magmaDouble_ptr dD = param.dA;
    magma_int_t lddd          = param.ldda; 
    magmaDouble_ptr dA = param.dB;
    magma_int_t ldda          = param.lddb;

    int info = 0;
    if ( uplo != MagmaLower && uplo != MagmaUpper && uplo != MagmaFull )
        info = -1;
    else if ( m < 0 )
        info = -2;
    else if ( n < 0 )
        info = -3;
    else if ( ldda < max(1,m) )
        info = -5;
    
    if (info != 0) {
        magma_xerbla( __func__, -(info) );
        return;  //info;
    }
    
    int grid = (m + NB - 1)/NB;
    int threads = NB ;
    
    if (uplo == MagmaLower) {
        dlascl_diag_lower( grid, threads, m, n, dD, lddd, dA, ldda);
    }
    else if (uplo == MagmaUpper) {
        dlascl_diag_upper( grid, threads, m, n, dD, lddd, dA, ldda);
    }
}

