/*
    -- micMAGMA (version 1.3.1) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date January 2015
 
       @author Mark Gates
       @author Stan Tomov

       @precisions normal z -> s d c
*/

#include <stdlib.h>
#include <stdio.h>

#include "magma.h"
#include "magmablas.h"
#include "error.h"

#ifdef HAVE_MIC

#include "magma_mic_bindings.h"

typedef struct {
    magma_int_t n;
    magma_int_t elemSize;
    void const* dx_src;
    magma_int_t incx;
    void*      dy_dst;
    magma_int_t incy;
} magma_mic_vcopy_param;

typedef struct {
    magma_int_t m;
    magma_int_t n;
    magma_int_t elemSize;
    void const* dA_src;
    magma_int_t lda;
    void*       dB_dst;
    magma_int_t ldb;
} magma_mic_mcopy_param;


// generic, type-independent routines to copy data.
// type-safe versions which avoid the user needing sizeof(...) are in copy_[sdcz].cpp

// ========================================
// copying vectors
extern "C"
void magma_setvector_internal(
    magma_int_t n, magma_int_t elemSize,
    void const* hx_src, magma_int_t incx,
    void*       dy_dst, magma_int_t incy,
    magma_queue_t queue,
    const char* func, const char* file, int line )
{
    magma_event_t event;
    magma_mic_synchronize_kernels();
    magma_mic_memcpy2D_async(
        dy_dst, incy*elemSize,
        hx_src, incx*elemSize,
        elemSize, n,
        WRITETO_DEV, queue, &event );
    magma_mic_fence_wait( event );
    //check_xerror( status, func, file, line );
}

// --------------------
extern "C"
void magma_getvector_internal(
    magma_int_t n, magma_int_t elemSize,
    void const* dx_src, magma_int_t incx,
    void*       hy_dst, magma_int_t incy,
    magma_queue_t queue,
    const char* func, const char* file, int line )
{
    magma_event_t event;
    magma_mic_synchronize_kernels();
    magma_mic_memcpy2D_async(
        hy_dst, incy*elemSize,
        dx_src, incx*elemSize,
        elemSize, n,
        READFROM_DEV, queue, &event );
    magma_mic_fence_wait( event );
    //check_xerror( status, func, file, line );
}

// --------------------
extern "C"
void magma_setvector_async_internal(
    magma_int_t n, magma_int_t elemSize,
    void const* hx_src, magma_int_t incx,
    void*       dy_dst, magma_int_t incy,
    magma_queue_t queue,
    const char* func, const char* file, int line )
{
    magma_mic_memcpy2D_async(
        dy_dst, incy*elemSize,
        hx_src, incx*elemSize,
        elemSize, n,
        WRITETO_DEV, queue, NULL );
    //check_xerror( status, func, file, line );
}

// --------------------
extern "C"
void magma_getvector_async_internal(
    magma_int_t n, magma_int_t elemSize,
    void const* dx_src, magma_int_t incx,
    void*       hy_dst, magma_int_t incy,
    magma_queue_t queue,
    const char* func, const char* file, int line )
{
    magma_mic_memcpy2D_async(
        hy_dst, incy*elemSize,
        dx_src, incx*elemSize,
        elemSize, n,
        READFROM_DEV, queue, NULL );
    //check_xerror( status, func, file, line );
}

// --------------------
// TODO compare performance with cublasZcopy BLAS function.
// But this implementation can handle any element size, not just [sdcz] precisions.
extern "C"
void magma_copyvector_internal(
    magma_int_t n, magma_int_t elemSize,
    void const* dx_src, magma_int_t incx,
    void*       dy_dst, magma_int_t incy,
    magma_queue_t queue,
    const char* func, const char* file, int line )
{
    int err;
    magma_mic_vcopy_param param = {n, elemSize, dx_src, incx, dy_dst, incy};

    int control_msg = magma_mic_VCOPY;

    if ((err = scif_send(gEpd, &control_msg, sizeof(control_msg), 1)) <= 0) {
        err = errno;
        printf("scif_send failed with err %d\n", errno);
        fflush(stdout);
    }

    if ((err = scif_send(gEpd, &param, sizeof(param), 1)) <= 0) {
        err = errno;
        printf("scif_send failed with err %d\n", errno);
        fflush(stdout);
    }
    //check_xerror( status, func, file, line );
}

// --------------------
extern "C"
void magma_copyvector_async_internal(
    magma_int_t n, magma_int_t elemSize,
    void const* dx_src, magma_int_t incx,
    void*       dy_dst, magma_int_t incy,
    magma_queue_t queue,
    const char* func, const char* file, int line )
{
    magma_copyvector_internal(n, elemSize, dx_src, incx,
                              dy_dst, incy, queue, func, file, line); 
    //check_xerror( status, func, file, line );
}


// ========================================
// copying sub-matrices (contiguous columns)
extern "C"
void magma_setmatrix_internal(
    magma_int_t m, magma_int_t n, magma_int_t elemSize,
    void const* hA_src, magma_int_t lda,
    void*       dB_dst, magma_int_t ldb,
    magma_queue_t queue,
    const char* func, const char* file, int line )
{
    printf( "%s not implemented\n", __func__ );
    //check_xerror( status, func, file, line );
}

// --------------------
extern "C"
void magma_getmatrix_internal(
    magma_int_t m, magma_int_t n, magma_int_t elemSize,
    void const* dA_src, magma_int_t lda,
    void*       hB_dst, magma_int_t ldb,
    magma_queue_t queue,
    const char* func, const char* file, int line )
{
    printf( "%s not implemented\n", __func__ );
    //check_xerror( status, func, file, line );
}

// --------------------
extern "C"
void magma_setmatrix_async_internal(
    magma_int_t m, magma_int_t n, magma_int_t elemSize,
    void const* hA_src, magma_int_t lda,
    void*       dB_dst, magma_int_t ldb,
    magma_queue_t queue,
    const char* func, const char* file, int line )
{
    printf( "%s not implemented\n", __func__ );
    //check_xerror( status, func, file, line );
}

// --------------------
extern "C"
void magma_getmatrix_async_internal(
    magma_int_t m, magma_int_t n, magma_int_t elemSize,
    void const* dA_src, magma_int_t lda,
    void*       hB_dst, magma_int_t ldb,
    magma_queue_t queue,
    const char* func, const char* file, int line )
{
    printf( "%s not implemented\n", __func__ );
    //check_xerror( status, func, file, line );
}

// --------------------
extern "C"
void magma_copymatrix_internal(
    magma_int_t m, magma_int_t n, magma_int_t elemSize,
    void const* dA_src, magma_int_t lda,
    void*       dB_dst, magma_int_t ldb,
    magma_queue_t queue,
    const char* func, const char* file, int line )
{
    int err;
    magma_mic_mcopy_param param = {m, n, elemSize, dA_src, lda, dB_dst, ldb};

    int control_msg = magma_mic_MCOPY;

    if ((err = scif_send(gEpd, &control_msg, sizeof(control_msg), 1)) <= 0) {
        err = errno;
        printf("scif_send failed with err %d\n", errno);
        fflush(stdout);
    }

    if ((err = scif_send(gEpd, &param, sizeof(param), 1)) <= 0) {
        err = errno;
        printf("scif_send failed with err %d\n", errno);
        fflush(stdout);
    }
    //check_xerror( status, func, file, line );
}

// --------------------
extern "C"
void magma_copymatrix_async_internal(
    magma_int_t m, magma_int_t n, magma_int_t elemSize,
    void const* dA_src, magma_int_t lda,
    void*       dB_dst, magma_int_t ldb,
    magma_queue_t queue,
    const char* func, const char* file, int line )
{
    magma_copymatrix_internal(m, n, elemSize,
                              dA_src, lda, dB_dst, ldb,
                              queue, func, file, line );
    //check_xerror( status, func, file, line );
}

#endif // HAVE_MIC
