/*
    -- MAGMA (version 1.3.1) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date January 2015

       @generated from zgerbt_mic.cpp normal z -> d, Fri Jan 30 18:21:52 2015
       @author Adrien REMY
*/
#include "common_magma.h"
#include "../testing/testings.h"
#define PRECISION_d
#define REAL



void 
init_butterfly(double* u, double* v, magma_int_t n)
{

    magma_int_t idx;
    double u1, v1;
    for (idx=0; idx<n; idx++){
        u1 = exp((((rand() * 1.0)/RAND_MAX)-0.5)/10);
        v1 = exp((((rand() * 1.0)/RAND_MAX)-0.5)/10);
        u[idx] = MAGMA_D_MAKE(u1,u1);

        v[idx] = MAGMA_D_MAKE(v1,v1);
    }
}


/**
    Purpose
    -------
    Solves a system of linear equations
       A * X = B
    where A is a general N-by-N matrix and X and B are N-by-nrhs matrices.
    Random Butterfly Tranformation is applied on A and B, then 
    the LU decomposition with no pivoting is
    used to factor A as
       A = L * U,
    where L is unit lower triangular, and U is
    upper triangular.  The factored form of A is then used to solve the
    system of equations A * X = B.

    Arguments
    ---------
    
    @param[in]
    gen     magma_bool_t
     -         = MagmaTrue:     new matrices are generated for U and V
     -         = MagmaFalse:    matrices U and V given as parameter are used
    
    @param[in]
    n       INTEGER
            The order of the matrix A.  N >= 0.

    @param[in]
    nrhs    INTEGER
            The number of right hand sides, i.e., the number of columns
            of the matrix B.  nrhs >= 0.

    @param[in,out]
    dA       DOUBLE_PRECISION array, dimension (LDA,N).
            On entry, the M-by-N matrix to be factored.
            On exit, the factors L and U from the factorization
            A = P*L*U; the unit diagonal elements of L are not stored.

    @param[in]
    ldda     INTEGER
            The leading dimension of the array A.  LDA >= max(1,N).

    
    @param[in,out]
    dB       DOUBLE_PRECISION array, dimension (LDB,nrhs)
            On entry, the right hand side matrix B.
            On exit, the solution matrix X.

    @param[in]
    lddb     INTEGER
            The leading dimension of the array B.  LDB >= max(1,N).

    @param[in,out]
    U        DOUBLE_PRECISION array, dimension (2,N)
            Random butterfly matrix, if gen = 'N' U is generated and returned as output
        else we use U given as input
        CPU memory
    @param[in,out]
    V        DOUBLE_PRECISION array, dimension (2,N)
            Random butterfly matrix, if gen = 'N' V is generated and returned as output
        else we use U given as input
        CPU memory

 ********************************************************************/



extern "C" 
magma_int_t 
magma_dgerbt_mic(
    magma_bool_t gen, magma_int_t n, magma_int_t nrhs, 
    double *dA, magma_int_t ldda, 
    double *dB, magma_int_t lddb, 
    double *U, double *V, 
    magma_int_t *info, magma_queue_t queue)
{
    magma_int_t ret;

    /* Function Body */
    *info = 0;
    if (n < 0) {
        *info = -2;
    } else if (nrhs < 0) {
        *info = -3;
    } else if (ldda < max(1,n)) {
        *info = -5;
    } else if (lddb < max(1,n)) {
        *info = -7;
    }
    if (*info != 0) {
        magma_xerbla( __func__, -(ret) );

        return *info;
    }

    /* Quick return if possible */
    if (nrhs == 0 || n == 0)
        return *info;


    

    magma_int_t n2;

    n2 = n*n;


    double *d_u, *d_v;

    /* Allocate memory for the buterfly matrices */
    if (MAGMA_SUCCESS != magma_dmalloc( &d_u, 2*n )) {
        *info = MAGMA_ERR_DEVICE_ALLOC;
        return *info;
    }
    if (MAGMA_SUCCESS != magma_dmalloc( &d_v, 2*n )) {
        *info = MAGMA_ERR_DEVICE_ALLOC;
        return *info;
    }

    /* Initialize Butterfly matrix on the CPU*/
    if (gen == MagmaTrue)
        init_butterfly(U,V,2*n);


    /* Copy the butterfly to the GPU */
    
    magma_dsetvector( 2*n, U, 1, d_u, 1, queue);
    magma_dsetvector( 2*n, V, 1, d_v, 1, queue);

    /* Perform Partial Random Butterfly Transformation on the GPU*/

    magma_dprbt(n, dA, ldda, d_u, d_v, queue);


    /* Compute U^T.b on the GPU*/
    for(int i= 0; i < nrhs; i++)
        magma_dprbt_mtv(n, d_u, dB+(i*lddb), queue);

    magma_free( d_u );
    magma_free( d_v );

    return *info;

}


