/*
    -- micMAGMA (version 1.3.1) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date January 2015

        @authors Azzam Haidar
        @authors Yulu Jia
        @authors Stan Tomov
 
        @precisions normal z -> s d c

*/
#include "common_magma.h"
//#define ENABLE_TIMER

#define A(i, j)  (a + (j)*lda  + (i))
#define dA(i, j) (da+ (j)*ldda + (i))

extern "C" magma_int_t
magma_zgebrd(magma_int_t m, magma_int_t n,
             magmaDoubleComplex *a, magma_int_t lda, double *d, double *e,
             magmaDoubleComplex *tauq, magmaDoubleComplex *taup,
             magmaDoubleComplex *work, magma_int_t lwork,
             magma_int_t *info, magma_queue_t queue)
{
/*  -- MAGMA (version 1.3.1) --
       Univ. of Tennessee, Knoxville
       Univ. of California, Berkeley
       Univ. of Colorado, Denver
       @date January 2015

    Purpose
    =======
    ZGEBRD reduces a general complex M-by-N matrix A to upper or lower
    bidiagonal form B by an orthogonal transformation: Q**H * A * P = B.

    If m >= n, B is upper bidiagonal; if m < n, B is lower bidiagonal.

    Arguments
    =========
    M       (input) INTEGER
            The number of rows in the matrix A.  M >= 0.

    N       (input) INTEGER
            The number of columns in the matrix A.  N >= 0.

    A       (input/output) COMPLEX_16 array, dimension (LDA,N)
            On entry, the M-by-N general matrix to be reduced.
            On exit,
            if m >= n, the diagonal and the first superdiagonal are
              overwritten with the upper bidiagonal matrix B; the
              elements below the diagonal, with the array TAUQ, represent
              the orthogonal matrix Q as a product of elementary
              reflectors, and the elements above the first superdiagonal,
              with the array TAUP, represent the orthogonal matrix P as
              a product of elementary reflectors;
            if m < n, the diagonal and the first subdiagonal are
              overwritten with the lower bidiagonal matrix B; the
              elements below the first subdiagonal, with the array TAUQ,
              represent the orthogonal matrix Q as a product of
              elementary reflectors, and the elements above the diagonal,
              with the array TAUP, represent the orthogonal matrix P as
              a product of elementary reflectors.
            See Further Details.

    LDA     (input) INTEGER
            The leading dimension of the array A.  LDA >= max(1,M).

    D       (output) double precision array, dimension (min(M,N))
            The diagonal elements of the bidiagonal matrix B:
            D(i) = A(i,i).

    E       (output) double precision array, dimension (min(M,N)-1)
            The off-diagonal elements of the bidiagonal matrix B:
            if m >= n, E(i) = A(i,i+1) for i = 1,2,...,n-1;
            if m < n, E(i) = A(i+1,i) for i = 1,2,...,m-1.

    TAUQ    (output) COMPLEX_16 array dimension (min(M,N))
            The scalar factors of the elementary reflectors which
            represent the orthogonal matrix Q. See Further Details.

    TAUP    (output) COMPLEX_16 array, dimension (min(M,N))
            The scalar factors of the elementary reflectors which
            represent the orthogonal matrix P. See Further Details.

    WORK    (workspace/output) COMPLEX_16 array, dimension (MAX(1,LWORK))
            On exit, if INFO = 0, WORK[0] returns the optimal LWORK.

    LWORK   (input) INTEGER
            The length of the array WORK.  LWORK >= (M+N)*NB, where NB
            is the optimal blocksize.

            If LWORK = -1, then a workspace query is assumed; the routine
            only calculates the optimal size of the WORK array, returns
            this value as the first entry of the WORK array, and no error
            message related to LWORK is issued by XERBLA.

    INFO    (output) INTEGER
            = 0:  successful exit
            < 0:  if INFO = -i, the i-th argument had an illegal value.

    Further Details
    ===============
    The matrices Q and P are represented as products of elementary
    reflectors:

    If m >= n,
       Q = H(1) H(2) . . . H(n)  and  P = G(1) G(2) . . . G(n-1)
    Each H(i) and G(i) has the form:
       H(i) = I - tauq * v * v'  and G(i) = I - taup * u * u'
    where tauq and taup are complex scalars, and v and u are complex vectors;
    v(1:i-1) = 0, v(i) = 1, and v(i+1:m) is stored on exit in A(i+1:m,i);
    u(1:i) = 0, u(i+1) = 1, and u(i+2:n) is stored on exit in A(i,i+2:n);
    tauq is stored in TAUQ(i) and taup in TAUP(i).

    If m < n,
       Q = H(1) H(2) . . . H(m-1)  and  P = G(1) G(2) . . . G(m)
    Each H(i) and G(i) has the form:
       H(i) = I - tauq * v * v'  and G(i) = I - taup * u * u'
    where tauq and taup are complex scalars, and v and u are complex vectors;
    v(1:i) = 0, v(i+1) = 1, and v(i+2:m) is stored on exit in A(i+2:m,i);
    u(1:i-1) = 0, u(i) = 1, and u(i+1:n) is stored on exit in A(i,i+1:n);
    tauq is stored in TAUQ(i) and taup in TAUP(i).

    The contents of A on exit are illustrated by the following examples:

    m = 6 and n = 5 (m > n):          m = 5 and n = 6 (m < n):

      (  d   e   u1  u1  u1 )           (  d   u1  u1  u1  u1  u1 )
      (  v1  d   e   u2  u2 )           (  e   d   u2  u2  u2  u2 )
      (  v1  v2  d   e   u3 )           (  v1  e   d   u3  u3  u3 )
      (  v1  v2  v3  d   e  )           (  v1  v2  e   d   u4  u4 )
      (  v1  v2  v3  v4  d  )           (  v1  v2  v3  e   d   u5 )
      (  v1  v2  v3  v4  v5 )

    where d and e denote diagonal and off-diagonal elements of B, vi
    denotes an element of the vector defining H(i), and ui an element of
    the vector defining G(i).
    =====================================================================    */
    #if defined (ENABLE_TIMER)
    double tbrd, tmtc, tgem, tctm, ttot, talcpy;
    #endif

    magmaDoubleComplex c_neg_one = MAGMA_Z_NEG_ONE;
    magmaDoubleComplex c_one     = MAGMA_Z_ONE;
    magmaDoubleComplex *da, *dwork;

    magma_int_t ncol, nrow, upd_row, upd_col, jmax, nb, ldda;
    magma_int_t fallback_cpu, labrd_fallback_cpu;
    magma_int_t i, j;
    magma_int_t iinfo;

    magma_int_t minmn;
    magma_int_t ldwrkx, ldwrky, lwkopt;
    magma_int_t lquery;

    nb   = magma_get_zgebrd_nb(n);
    ldda    = ((m+31)/32)*32;
    //ldda = m;
    /* Set the block/unblock crossover point fallback_cpu. */
    // NOTE fallback_cpu should be at least nb
    fallback_cpu = nb;
    fallback_cpu = 4800;
    labrd_fallback_cpu = 4080;


    lwkopt = (m + n) * nb;
    work[0] = MAGMA_Z_MAKE( lwkopt, 0. );
    lquery = (lwork == -1);
    
    /* Check arguments */
    *info = 0;
    if (m < 0) {
        *info = -1;
    } else if (n < 0) {
        *info = -2;
    } else if (lda < max(1,m)) {
        *info = -4;
    } else if (lwork < lwkopt && (! lquery) ) {
        *info = -10;
    }
    if (*info < 0) {
        magma_xerbla( __func__, -(*info) );
        return *info;
    }
    else if (lquery)
        return *info;

    /* Quick return if possible */
    minmn = min(m,n);
    if (minmn == 0) {
        work[0] = c_one;
        return *info;
    }

    #if defined (ENABLE_TIMER)
    talcpy = magma_wtime();
    #endif

    if (MAGMA_SUCCESS != magma_zmalloc( &da, n*ldda + (m + n)*nb )) {
        fprintf (stderr, "!!!! device memory allocation error in zgebrd\n" );
        *info = MAGMA_ERR_DEVICE_ALLOC;
        return *info;
    }
    dwork = da + (n)*ldda;

    ldwrkx = m;
    ldwrky = n;

    /* Copy the matrix to the GPU */
    if (minmn - fallback_cpu >= 1) {
        magma_zsetmatrix( m, n, a, 0, lda, da, 0, ldda, queue );
    }
    #if defined (ENABLE_TIMER)
    magma_kernel_sync();
    talcpy = magma_wtime() - talcpy;
    #endif

    #if defined (ENABLE_TIMER)
    ttot = magma_wtime();
    #endif

    #if defined (ENABLE_TIMER)
    printf("\n\n%8s %8s %8s %8s %8s %8s %8s %8s %8s %8s %8s %8s %8s %8s %8s\n",
                  "ctm1","gmv1","mtc1","ctm2","gmv2","mtc2", "tcopy", "tcpu", "tpanel", "i", "row", "nb", "mtc", "GEMM", "ctm");
    printf("     ===========================================================================================================================================\n");
    //printf("\n\n %8s %8s %8s %8s %8s %8s %8s %8s %8s %8s \n", 
    //              "i", "row", "nb", "panel",  "mtc", "CPU", "MIC", "GEMM", "ctm", "frst cpy");
    //printf("     ====================================================================================================\n");
    #endif
    for (i=0; i < (minmn - fallback_cpu); i += nb) {

        /*  Reduce rows and columns i:i+nb-1 to bidiagonal form and return
            the matrices X and Y which are needed to update the unreduced
            part of the matrix */
        nrow = m - i;
        ncol = n - i;

        if( nrow > labrd_fallback_cpu ){

        #if defined (ENABLE_TIMER)
        tmtc = magma_wtime();
        #endif

        /*   Get the current panel (no need for the 1st iteration) */
        if ( i > 0 ) {
            magma_zgetmatrix( nrow, nb, dA(i, i), 0, ldda, A( i, i), 0, lda, queue );
            magma_zgetmatrix( nb, ncol - nb,
                              dA(i, i+nb), 0, ldda,
                              A( i, i+nb), 0, lda, queue );
        }

        #if defined (ENABLE_TIMER)
        magma_kernel_sync();
        tmtc = magma_wtime() - tmtc;
        #endif
        
        #if defined (ENABLE_TIMER)
        tbrd = magma_wtime();
        #endif
        magma_zlabrd_mic(nrow, ncol, nb,
                         A(i, i),          lda,    dA(i, i),          0,   ldda,
                         d+i, e+i, tauq+i, taup+i,
                         work,             ldwrkx, dwork,             0, ldwrkx,  // x, dx
                         work+(ldwrkx*nb), ldwrky, dwork+(ldwrkx*nb), 0, ldwrky, // y, dy
                         i, queue );
        #if defined (ENABLE_TIMER)
        magma_kernel_sync();
        tbrd = magma_wtime() - tbrd;
        #endif
        }
        else{
            #if defined (ENABLE_TIMER)
            tmtc = magma_wtime();
            #endif
            magma_zgetmatrix( nrow, ncol, dA(i, i), 0, ldda, A( i, i), 0, lda, queue );
            #if defined (ENABLE_TIMER)
            tmtc = magma_wtime() - tmtc;
            #endif
            #if defined (ENABLE_TIMER)
            tbrd = magma_wtime();
            #endif
	        lapackf77_zlabrd( &nrow, &ncol, &nb, 
			                A(i, i), &lda, 
			                d+i, e+i, tauq+i, taup+i,
                            work,             &ldwrkx,
                            work+(ldwrkx*nb), &ldwrky);
            #if defined (ENABLE_TIMER)
            tbrd = magma_wtime() - tbrd;
            #endif
	        // send back the panels only
            #if defined (ENABLE_TIMER)
            tctm = magma_wtime();
            #endif
            magma_zsetmatrix( nrow, nb,    A( i, i), 0, lda, dA(i, i), 0, ldda, queue );
            magma_zsetmatrix( nb, ncol-nb, A( i, i+nb), 0, lda, dA(i, i+nb), 0, ldda, queue );
            #if defined (ENABLE_TIMER)
            tctm = magma_wtime() - tctm;
            #endif
        }
        /*  Update the trailing submatrix A(i+nb:m,i+nb:n), using an update
            of the form  A := A - V*Y' - X*U' */
        upd_row = m - i - nb;
        upd_col = n - i - nb;

        // Send Y back to the GPU
        #if defined (ENABLE_TIMER)
        tctm = magma_wtime();
        #endif
        magma_zsetmatrix( upd_row, nb, work  + nb, 0, ldwrkx, dwork + nb, 0, ldwrkx, queue );
        magma_zsetmatrix( upd_col, nb,
                          work  + (ldwrkx+1)*nb, 0, ldwrky,
                          dwork + (ldwrkx+1)*nb, 0, ldwrky, queue );
        #if defined (ENABLE_TIMER)
        magma_kernel_sync();
        tctm = magma_wtime() - tctm;
        #endif

        #if defined (ENABLE_TIMER)
        tgem = magma_wtime();
        #endif
        magma_zgemm( MagmaNoTrans, MagmaConjTrans,
                     upd_row, upd_col, nb,
                     c_neg_one, dA(i+nb, i   ),      0, ldda,
                                dwork+(ldwrkx+1)*nb, 0, ldwrky,
                     c_one,     dA(i+nb, i+nb),      0, ldda, queue);

        magma_zgemm( MagmaNoTrans, MagmaNoTrans,
                     upd_row, upd_col, nb,
                     c_neg_one, dwork+nb,         0, ldwrkx,
                                dA( i,    i+nb ), 0, ldda,
                     c_one,     dA( i+nb, i+nb ), 0, ldda, queue);

        #if defined (ENABLE_TIMER)
        magma_kernel_sync();
        tgem = magma_wtime() - tgem;
        #endif
        /* Copy diagonal and off-diagonal elements of B back into A */
        if (m >= n) {
            jmax = i + nb;
            for (j = i; j < jmax; ++j) {
                *A(j, j  ) = MAGMA_Z_MAKE( d[j], 0. );
                *A(j, j+1) = MAGMA_Z_MAKE( e[j], 0. );
            }
        } else {
            jmax = i + nb;
            for (j = i; j < jmax; ++j) {
                *A(j,   j ) = MAGMA_Z_MAKE( d[j], 0. );
                *A(j+1, j ) = MAGMA_Z_MAKE( e[j], 0. );
            }
        }
        #if defined (ENABLE_TIMER)
        printf("%8d %8d %8d %8.2lf %8.2lf %8.2lf\n",
                               i, upd_row, nb, tmtc*1000, tgem * 1000, tctm * 1000);
        #endif
    }

    /* Use unblocked code to reduce the remainder of the matrix */
    nrow = m - i;
    ncol = n - i;

    if ( 0 < minmn - fallback_cpu ) {
        magma_zgetmatrix( nrow, ncol, dA(i, i), 0, ldda, A(i, i), 0, lda, queue );
    }
    
    lapackf77_zgebrd( &nrow, &ncol,
                      A(i, i), &lda, d+i, e+i,
                      tauq+i, taup+i, work, &lwork, &iinfo);
    work[0] = MAGMA_Z_MAKE( lwkopt, 0. );

    magma_free( da );

    #if defined (ENABLE_TIMER)
    magma_kernel_sync();
    ttot = magma_wtime() - ttot;
    printf("\n total: %8.1lf    alloc+first copy: %8.1lf \n", ttot, talcpy);
    #endif
    return *info;
} /* zgebrd */
